@@ -1,4 +1,3 @@
-// |reftest| shell-option(--enable-float16array)
 // Copyright (C) 2016 the V8 project authors. All rights reserved.
 // This code is governed by the BSD license found in the LICENSE file.
 /*---
@@ -23,8 +22,10 @@ var TypedArrayPrototype = TypedArray.prototype;
 
 assert.sameValue(TypedArrayPrototype.toString, Array.prototype.toString);
 
-verifyNotEnumerable(TypedArrayPrototype, 'toString');
-verifyWritable(TypedArrayPrototype, 'toString');
-verifyConfigurable(TypedArrayPrototype, 'toString');
+verifyProperty(TypedArrayPrototype, 'toString', {
+  writable: true,
+  enumerable: false,
+  configurable: true
+});
 
 reportCompare(0, 0);
