@@ -228,11 +228,12 @@ enum class LimitsFlags {
 };
 
 enum class LimitsMask {
-  Table = uint8_t(LimitsFlags::HasMaximum),
 #ifdef ENABLE_WASM_MEMORY64
+  Table = uint8_t(LimitsFlags::HasMaximum) | uint8_t(LimitsFlags::IsI64),
   Memory = uint8_t(LimitsFlags::HasMaximum) | uint8_t(LimitsFlags::IsShared) |
            uint8_t(LimitsFlags::IsI64),
 #else
+  Table = uint8_t(LimitsFlags::HasMaximum),
   Memory = uint8_t(LimitsFlags::HasMaximum) | uint8_t(LimitsFlags::IsShared),
 #endif
 };
@@ -832,8 +833,8 @@ enum class SimdOp {
   F64x2RelaxedMin = 0x10f,
   F64x2RelaxedMax = 0x110,
   I16x8RelaxedQ15MulrS = 0x111,
-  I16x8DotI8x16I7x16S = 0x112,
-  I32x4DotI8x16I7x16AddS = 0x113,
+  I16x8RelaxedDotI8x16I7x16S = 0x112,
+  I32x4RelaxedDotI8x16I7x16AddS = 0x113,
 
   // Reserved for Relaxed SIMD = 0x114-0x12f
 
@@ -878,8 +879,8 @@ enum class MiscOp {
 
 // Opcodes from threads proposal as of June 30, 2017
 enum class ThreadOp {
-  // Wait and wake
-  Wake = 0x00,
+  // Wait and notify
+  Notify = 0x00,
   I32Wait = 0x01,
   I64Wait = 0x02,
   Fence = 0x03,
@@ -962,14 +963,16 @@ enum class ThreadOp {
 };
 
 enum class BuiltinModuleFuncId {
+  None = 0,
+
 // ------------------------------------------------------------------------
 // These are part/suffix of the MozOp::CallBuiltinModuleFunc operators that are
 // emitted internally when compiling intrinsic modules and are rejected by wasm
 // validation.
 // See wasm/WasmBuiltinModule.yaml for the list.
-#define VISIT_BUILTIN_FUNC(op, export, sa_name, abitype, entry, has_memory, \
-                           idx)                                             \
-  op = idx,  // NOLINT
+#define VISIT_BUILTIN_FUNC(op, export, sa_name, abitype, entry, uses_memory, \
+                           inline_op, idx)                                   \
+  op = idx + 1,  // NOLINT
   FOR_EACH_BUILTIN_MODULE_FUNC(VISIT_BUILTIN_FUNC)
 #undef VISIT_BUILTIN_FUNC
 
@@ -977,26 +980,26 @@ enum class BuiltinModuleFuncId {
   Limit
 };
 
+enum class BuiltinInlineOp {
+  None = 0,
+
+  StringCast,
+  StringTest,
+  StringLength,
+
+  // Op limit.
+  Limit
+};
+
 enum class BuiltinModuleId {
   SelfTest = 0,
   IntGemm,
   JSString,
+
+  // Not technically a builtin module, but it uses most of the same machinery.
+  JSStringConstants,
 };
 
-struct BuiltinModuleIds {
-  BuiltinModuleIds() = default;
-
-  bool selfTest = false;
-  bool intGemm = false;
-  bool jsString = false;
-
-  bool hasNone() const { return !selfTest && !intGemm && !jsString; }
-
-  WASM_CHECK_CACHEABLE_POD(selfTest, intGemm, jsString)
-};
-
-WASM_DECLARE_CACHEABLE_POD(BuiltinModuleIds)
-
 enum class StackSwitchKind {
   SwitchToSuspendable,
   SwitchToMain,
@@ -1073,8 +1076,19 @@ struct OpBytes {
     b0 = uint16_t(x);
     b1 = 0;
   }
+  OpBytes(uint16_t b0, uint16_t b1) : b0(b0), b1(b1) {}
   OpBytes() = default;
 
+  uint32_t toPacked() const {
+    // In practice all of our secondary bytecodes are actually 16-bit right now.
+    MOZ_RELEASE_ASSERT(b1 <= UINT16_MAX);
+    return b0 | (b1 << 16);
+  }
+
+  static OpBytes fromPacked(uint32_t packed) {
+    return OpBytes(packed & 0xFFFF, packed >> 16);
+  }
+
   // Whether this opcode should have a breakpoint site inserted directly before
   // the opcode in baseline when debugging. We use this as a heuristic to
   // reduce the number of breakpoint sites.
@@ -1106,6 +1120,9 @@ struct OpBytes {
         return true;
     }
   }
+
+  // Defined in WasmOpIter.cpp
+  const char* toString() const;
 };
 
 static const char NameSectionName[] = "name";
@@ -1136,21 +1153,22 @@ static const unsigned MaxTags = 1000000;
 static const unsigned MaxFuncs = 1000000;
 static const unsigned MaxTables = 100000;
 static const unsigned MaxMemories = 100000;
-static const unsigned MaxImports = 100000;
-static const unsigned MaxExports = 100000;
+static const unsigned MaxImports = 1000000;
+static const unsigned MaxExports = 1000000;
 static const unsigned MaxGlobals = 1000000;
 static const unsigned MaxDataSegments = 100000;
 static const unsigned MaxDataSegmentLengthPages = 16384;
 static const unsigned MaxElemSegments = 10000000;
 static const unsigned MaxElemSegmentLength = 10000000;
-static const unsigned MaxTableLimitField = UINT32_MAX;
-static const unsigned MaxTableLength = 10000000;
+static const uint64_t MaxTable32ElemsValidation = UINT32_MAX;
+static const uint64_t MaxTable64ElemsValidation = UINT64_MAX;
+static const unsigned MaxTableElemsRuntime = 10000000;
 static const unsigned MaxLocals = 50000;
 static const unsigned MaxParams = 1000;
 static const unsigned MaxResults = 1000;
 static const unsigned MaxStructFields = 10000;
-static const uint64_t MaxMemory32LimitField = uint64_t(1) << 16;
-static const uint64_t MaxMemory64LimitField = uint64_t(1) << 48;
+static const uint64_t MaxMemory32PagesValidation = uint64_t(1) << 16;
+static const uint64_t MaxMemory64PagesValidation = uint64_t(1) << 48;
 static const unsigned MaxStringBytes = 100000;
 static const unsigned MaxModuleBytes = 1024 * 1024 * 1024;
 static const unsigned MaxFunctionBytes = 7654321;
@@ -1202,10 +1220,36 @@ static constexpr size_t SuspendableStackPlusRedZoneSize =
 
 static const unsigned MaxVarU32DecodedBytes = 5;
 
-// The CompileMode controls how compilation of a module is performed (notably,
-// how many times we compile it).
+// The CompileMode controls how compilation of a module is performed.
+enum class CompileMode {
+  // Compile the module just once using a given tier.
+  Once,
+  // Compile the module first with baseline, then eagerly launch an ion
+  // background compile to compile the module again.
+  EagerTiering,
+  // Compile the module first with baseline, then lazily compile functions with
+  // ion when they trigger a hotness threshold.
+  LazyTiering,
+};
 
-enum class CompileMode { Once, Tier1, Tier2 };
+// CompileState tracks where in the compilation process we are for a module.
+enum class CompileState {
+  // We're compiling the module using the 'once' mode.
+  Once,
+  // We're compiling the module using the eager tiering mode. We're
+  // currently compiling the first tier. The second tier task will be launched
+  // once we're done compiling the first tier.
+  EagerTier1,
+  // We're compiling the module using the eager tiering mode. We're now
+  // compiling the second tier.
+  EagerTier2,
+  // We're compiling the module eagerly using the lazy tiering mode. We're
+  // compiling the first tier.
+  LazyTier1,
+  // We're compiling the module eagerly using the lazy tiering strategy. We're
+  // compiling the second tier.
+  LazyTier2,
+};
 
 // Typed enum for whether debugging is enabled.
 
